#include "DownloadQueueModel.h"

DownloadQueueModel::DownloadQueueModel(QNNTPGrabGlue *glue, QObject *parent) :
    QAbstractItemModel(parent)
{
    rootItem = new DownloadQueueItem();

    loadDownloadQueue(glue);

    connect(glue, SIGNAL(fileDownloadStateUpdate(QString,QString,int,int,int,quint64,quint64,quint64,quint64)), SLOT(onFileDownloadStateUpdate(QString,QString,int,int,int,quint64,quint64,quint64,quint64)), Qt::QueuedConnection);
    connect(glue, SIGNAL(fileStateChanged(QString,QString,QString,QNNTPGrabGlue::TaskState,QNNTPGrabGlue::TaskState,quint64,quint64,quint64)), SLOT(onFileStateChanged(QString,QString,QString,QNNTPGrabGlue::TaskState,QNNTPGrabGlue::TaskState,quint64,quint64,quint64)), Qt::QueuedConnection);

    connect(glue, SIGNAL(fileAdded(QString,QString,QString,QDateTime,quint64,quint64,quint64,QNNTPGrabGlue::TaskState,int,QList<QString>)), SLOT(onFileAdded(QString,QString,QString,QDateTime,quint64,quint64,quint64,QNNTPGrabGlue::TaskState,int,QList<QString>)), Qt::QueuedConnection);
    connect(glue, SIGNAL(fileRemoved(QString,QString,quint64,quint64)), SLOT(onFileRemoved(QString,QString,quint64,quint64)), Qt::QueuedConnection);
    connect(glue, SIGNAL(fileMoved(QString,QString,QString,int,int)), SLOT(onFileMoved(QString,QString,QString,int,int)), Qt::QueuedConnection);

    connect(glue, SIGNAL(collectionAdded(QString,QString)), SLOT(onCollectionAdded(QString,QString)), Qt::QueuedConnection);
    connect(glue, SIGNAL(collectionRemoved(QString)), SLOT(onCollectionRemoved(QString)), Qt::QueuedConnection);
    connect(glue, SIGNAL(collectionModified(QString,QString)), SLOT(onCollectionModified(QString,QString)), Qt::QueuedConnection);
    connect(glue, SIGNAL(collectionMoved(QString,int,int)), SLOT(onCollectionMoved(QString,int,int)), Qt::QueuedConnection);

    connect(glue, SIGNAL(trafficMonitorUpdate(int,int,int,int,int,int,int,int,int,int,QDateTime,double)), SLOT(onTrafficMonitorUpdate(int,int,int,int,int,int,int,int,int,int,QDateTime,double)), Qt::QueuedConnection);
}

DownloadQueueModel::~DownloadQueueModel()
{
    delete rootItem;
}

int DownloadQueueModel::columnCount(const QModelIndex &parent) const
{
    return parent.isValid() && parent.column() != 0 ? 0 : DownloadQueueItem::NUM_FIELDS;
}

QVariant DownloadQueueModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid())
        return QVariant();

    if (role != Qt::DisplayRole)
        return QVariant();

    DownloadQueueItem *item = static_cast<DownloadQueueItem*>(index.internalPointer());

    return item->data(index.column());
}

Qt::ItemFlags DownloadQueueModel::flags(const QModelIndex &index) const
{
    if (!index.isValid())
        return 0;

    return Qt::ItemIsEnabled | Qt::ItemIsSelectable;
}

QVariant DownloadQueueModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (section) {
            case DownloadQueueItem::FIELD_STATE_STR:
                return tr("Progress");

            case DownloadQueueItem::FIELD_SUBJECT:
                return tr("Subject");

            case DownloadQueueItem::FIELD_POSTER:
                return tr("Poster");

            case DownloadQueueItem::FIELD_GROUPS:
                return tr("Groups");

            case DownloadQueueItem::FIELD_FILE_SIZE_STR:
                return tr("File size");

            case DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR:
                return tr("File size remaining");

            case DownloadQueueItem::FIELD_STAMP:
                return tr("Stamp");

            case DownloadQueueItem::FIELD_ESTIMATED_TIME_REMAINING_STR:
                return tr("Estimated time remaining");

            case DownloadQueueItem::FIELD_ESTIMATED_TIME_TO_FINISH_STR:
                return tr("Estimated time to finish");

            case DownloadQueueItem::FIELD_PROGRESS:
            case DownloadQueueItem::FIELD_FILE_SIZE:
            case DownloadQueueItem::FIELD_FILE_SIZE_REMAINING:
            case DownloadQueueItem::FIELD_REAL_FILENAME:
            default:
                break;
        }
    }

    return QVariant();
}

QModelIndex DownloadQueueModel::index(int row, int column, const QModelIndex &parent) const
{
    if (!hasIndex(row, column, parent))
        return QModelIndex();

    DownloadQueueItem *parentItem;

    if (!parent.isValid())
        parentItem = rootItem;
    else
        parentItem = static_cast<DownloadQueueItem*>(parent.internalPointer());

    DownloadQueueItem *childItem = parentItem->child(row);
    if (childItem)
        return createIndex(row, column, childItem);
    else
        return QModelIndex();
}

QModelIndex DownloadQueueModel::parent(const QModelIndex &index) const
{
    if (!index.isValid())
        return QModelIndex();

    if (DownloadQueueItem *childItem = itemForIndex(index)) {
        if (DownloadQueueItem *parentItem = childItem->parent()) {
            if (parentItem == rootItem)
                return QModelIndex();
            if (DownloadQueueItem *grandParentItem = parentItem->parent()) {
                int row = grandParentItem->rowOfChild(parentItem);
                return createIndex(row, 0, parentItem);
            }
        }
    }

    return QModelIndex();
}

DownloadQueueItem *DownloadQueueModel::itemForIndex(const QModelIndex &index) const
{
    if (index.isValid()) {
        if (DownloadQueueItem *item = static_cast<DownloadQueueItem*>(index.internalPointer())) {
            return item;
        }
    }

    return rootItem;
}

int DownloadQueueModel::rowCount(const QModelIndex &parent) const
{
    if (parent.isValid() && parent.column() != 0)
        return 0;

    DownloadQueueItem *parentItem = itemForIndex(parent);
    return parentItem ? parentItem->childCount() : 0;
}

void DownloadQueueModel::loadDownloadQueue(QNNTPGrabGlue *glue)
{
    Q_ASSERT(glue != NULL);

    QList<QNNTPGrabGlue::Collection> collections = glue->schedularGetAllTasks();

    for (int i = 0; i < collections.count(); i++) {
        QNNTPGrabGlue::Collection collection = collections[i];

        DownloadQueueItem *collection_item = new DownloadQueueItem(rootItem);

        for (int j = 0; j < collection.files.count(); j++) {
            QNNTPGrabGlue::File file = collection.files[j];

            DownloadQueueItem *file_item = new DownloadQueueItem(collection_item);

            file_item->subject = file.subject;
            file_item->poster = file.poster;
            file_item->size = file.file_size;
            file_item->sizeStr = QNNTPGrabGlue::utilsCalculateFileSize(file.file_size);
            file_item->sizeRemaining = file.file_size_remaining;
            file_item->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(file.file_size_remaining);
            file_item->stamp = file.stamp;
            file_item->realFilename = file.filename;

            for (int k = 0; k < file.groups.count(); k++) {
                if (file_item->groups.isEmpty()) {
                    file_item->groups = file.groups[k];
                } else {
                    file_item->groups += QString(",") + file.groups[k];
                }
            }

            if (file.status == QNNTPGrabGlue::TASK_STATE_SKIPPED) {
                file_item->progress = 0;
            } else {
                file_item->progress = calculateProgress(file.file_size, file.file_size_remaining);
            }

            switch(file.status) {
                case QNNTPGrabGlue::TASK_STATE_WAITING_FOR_DOWNLOAD:
                case QNNTPGrabGlue::TASK_STATE_DOWNLOADING:
                    file_item->stateStr = "";
                    break;

                case QNNTPGrabGlue::TASK_STATE_WAITING_FOR_DECODE:
                    file_item->stateStr = tr("Waiting");
                    break;

                case QNNTPGrabGlue::TASK_STATE_DECODING:
                    file_item->stateStr = tr("Decoding");
                    break;

                case QNNTPGrabGlue::TASK_STATE_FINISHED_COMPLETE:
                    file_item->stateStr = tr("Done");
                    break;

                case QNNTPGrabGlue::TASK_STATE_FINISHED_INCOMPLETE:
                    file_item->stateStr = tr("Incomplete");
                    break;

                case QNNTPGrabGlue::TASK_STATE_FINISHED_NO_PARTS_AVAIL:
                    file_item->stateStr = tr("Not available");
                    break;

                case QNNTPGrabGlue::TASK_STATE_SKIPPED:
                    file_item->stateStr = tr("Skipped");
                    break;

                default:
                    file_item->stateStr = "";
                    break;
            };

            collection_item->appendChild(file_item);
        }

        collection_item->progress = calculateProgress(collection.total_size, collection.total_size_remaining);
        collection_item->subject = collection.collection_name;
        collection_item->poster = collection.poster;
        collection_item->size = collection.total_size;
        collection_item->sizeStr = QNNTPGrabGlue::utilsCalculateFileSize(collection.total_size);
        collection_item->sizeRemaining = collection.total_size_remaining;
        collection_item->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(collection.total_size_remaining);
        collection_item->stamp = collection.files.at(0).stamp;

        rootItem->appendChild(collection_item);
    }
}

int DownloadQueueModel::calculateProgress(quint64 size, quint64 sizeRemaining)
{
    if (size == 0) {        /* Prevent a divide by zero */
        return 0;
    }

    int ret = (((double)size - sizeRemaining) / size) * 100;

    if (ret < 0 || ret > 100) {
        qDebug("Invalid progress detected, size = %lli, size_remaining = %lli", size, sizeRemaining);
        if (ret < 0) {
            ret = 0;
        }
    }

    return qMin(100, ret);
}

DownloadQueueItem *DownloadQueueModel::lookupCollectionItem(QString collection_name)
{
    if (!rootItem) {
        return NULL;
    }

    return rootItem->lookupSubject(collection_name);
}

DownloadQueueItem *DownloadQueueModel::lookupSubjectItem(DownloadQueueItem *collection_item, QString subject)
{
    if (!collection_item) {
        return NULL;
    }

    return collection_item->lookupSubject(subject);
}

void DownloadQueueModel::onFileDownloadStateUpdate(QString collection_name, QString subject, int num_parts_total, int num_parts_done,
                                                   int num_parts_failed, quint64 file_size, quint64 file_size_remaining,
                                                   quint64 total_size, quint64 total_size_remaining)
{
    Q_UNUSED(num_parts_total);
    Q_UNUSED(num_parts_done);
    Q_UNUSED(num_parts_failed);

    DownloadQueueItem *collection_node;
    DownloadQueueItem *subject_node;

    collection_node = lookupCollectionItem(collection_name);
    subject_node = lookupSubjectItem(collection_node, subject);
    if (subject_node == NULL) {
        qWarning("onFileDownloadStateUpdate: Subject '%s' from collection '%s' could not found", collection_name.toUtf8().data(), subject.toUtf8().data());
        return;
    }

    /* Don't continue processing this event when the file just got skipped */
    if (subject_node->stateStr == tr("Skipped")) {
        return;
    }

    QModelIndex idx;

    collection_node->progress = calculateProgress(total_size, total_size_remaining);
    collection_node->sizeRemaining = total_size_remaining;
    collection_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size_remaining);

    idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, collection_node);
    emit dataChanged(idx, idx);

    idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, collection_node);
    emit dataChanged(idx, idx);

    subject_node->progress = calculateProgress(file_size, file_size_remaining);
    subject_node->sizeRemaining = file_size_remaining;
    subject_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(file_size_remaining);

    idx = createIndex(subject_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, subject_node);
    emit dataChanged(idx, idx);

    idx = createIndex(subject_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, subject_node);
    emit dataChanged(idx, idx);
}

void DownloadQueueModel::onFileStateChanged(QString collection_name, QString subject, QString real_filename, QNNTPGrabGlue::TaskState old_state, QNNTPGrabGlue::TaskState new_state, quint64 file_size_remaining, quint64 total_size, quint64 total_size_remaining)
{
    Q_UNUSED(old_state);

    QModelIndex idx;
    DownloadQueueItem *collection_node;
    DownloadQueueItem *subject_node;

    collection_node = lookupCollectionItem(collection_name);
    subject_node = lookupSubjectItem(collection_node, subject);

    if (!subject_node) {
        qWarning("onFileStateChanged: Subject '%s' from collection '%s' could not be found", subject.toUtf8().data(), collection_name.toUtf8().data());
        return;
    }

    switch(new_state) {
        case QNNTPGrabGlue::TASK_STATE_WAITING_FOR_DOWNLOAD:
            /* The task was restarted */

            /* Update the row itself */
            subject_node->progress = 0;
            subject_node->stateStr = "";
            subject_node->sizeRemaining = file_size_remaining;
            subject_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(file_size_remaining);

            /* Update the parent */
            collection_node->progress = calculateProgress(total_size, total_size_remaining);
            collection_node->size = total_size;
            collection_node->sizeStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size);
            collection_node->sizeRemaining = total_size_remaining;
            collection_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size_remaining);

            idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, collection_node);
            emit dataChanged(idx, idx);

            idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, collection_node);
            emit dataChanged(idx, idx);

            break;
        case QNNTPGrabGlue::TASK_STATE_DOWNLOADING:
            subject_node->progress = 0;
            subject_node->stateStr = "";                /* The state message should be empty. In that case the value of the progress bar will be shown */
            break;
        case QNNTPGrabGlue::TASK_STATE_WAITING_FOR_DECODE:
            subject_node->progress = 100;
            subject_node->sizeRemaining = 0;
            subject_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(0);
            subject_node->stateStr = tr("Waiting");
            break;
        case QNNTPGrabGlue::TASK_STATE_DECODING:
            subject_node->progress = 100;
            subject_node->stateStr = tr("Decoding");
            break;
        case QNNTPGrabGlue::TASK_STATE_FINISHED_COMPLETE:
            Q_ASSERT(real_filename != NULL);

            subject_node->stateStr = tr("Done");
            subject_node->realFilename = real_filename;
            break;
        case QNNTPGrabGlue::TASK_STATE_FINISHED_INCOMPLETE:
            Q_ASSERT(real_filename != NULL);

            subject_node->stateStr = tr("Incomplete");
            subject_node->realFilename = real_filename;
            break;
        case QNNTPGrabGlue::TASK_STATE_FINISHED_NO_PARTS_AVAIL:
            subject_node->stateStr = tr("Not available");
            break;
        case QNNTPGrabGlue::TASK_STATE_SKIPPED:
            subject_node->stateStr = tr("Skipped");
            subject_node->sizeRemaining = 0;
            subject_node->sizeRemainingStr = "";

            /* Update the parent */
            collection_node->progress = calculateProgress(total_size, total_size_remaining);
            collection_node->size = total_size;
            collection_node->sizeStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size);
            collection_node->sizeRemaining = total_size_remaining;
            collection_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size_remaining);

            idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, collection_node);
            emit dataChanged(idx, idx);

            idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, collection_node);
            emit dataChanged(idx, idx);

            break;
    }

    idx = createIndex(subject_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, subject_node);
    emit dataChanged(idx, idx);

    idx = createIndex(subject_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, subject_node);
    emit dataChanged(idx, idx);

    emit requestColumnResize(DownloadQueueItem::FIELD_STATE_STR);
}

void DownloadQueueModel::onFileAdded(QString collection_name, QString subject, QString poster, QDateTime stamp, quint64 file_size, quint64 total_size, quint64 total_size_remaining, QNNTPGrabGlue::TaskState state, int num_parts, QList<QString> groups)
{
    Q_UNUSED(num_parts);

    DownloadQueueItem *collection_node = lookupCollectionItem(collection_name);
    if (collection_node == NULL) {
        qWarning("file added: Collection %s could not be found", collection_name.toUtf8().data());
        return;
    }

    beginInsertRows(createIndex(collection_node->row(), 0, collection_node), collection_node->childCount(), collection_node->childCount());

    DownloadQueueItem *file_node = new DownloadQueueItem(collection_node);
    file_node->subject = subject;
    file_node->poster = poster;
    file_node->size = file_size;
    file_node->sizeStr = QNNTPGrabGlue::utilsCalculateFileSize(file_size);
    file_node->sizeRemaining = file_size;
    file_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(file_size);
    file_node->stamp = stamp;

    int i;
    QString groups_str;
    for (i = 0; i < groups.count(); i++) {
        if (groups_str.isEmpty()) {
            groups_str = groups[i];
        } else {
            groups_str += QString(",") + groups[i];
        }
    }
    file_node->groups = groups_str;

    if (state == QNNTPGrabGlue::TASK_STATE_SKIPPED) {
        file_node->stateStr = tr("Skipped");
        file_node->sizeRemaining = 0;
        file_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(0);
    }

    /* Update the parent */
    collection_node->appendChild(file_node);

    endInsertRows();

    collection_node->size = total_size;
    collection_node->sizeStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size);
    collection_node->sizeRemaining = total_size_remaining;
    collection_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size_remaining);
    collection_node->progress = calculateProgress(total_size, total_size_remaining);

    /* Set the stamp if this is the first file of the collection */
    if (collection_node->childCount() == 1) {
        collection_node->stamp = stamp;
    }

    QModelIndex idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, collection_node);
    emit dataChanged(idx, idx);

    idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_STR, collection_node);
    emit dataChanged(idx, idx);

    idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, collection_node);
    emit dataChanged(idx, idx);

    emit requestColumnResize(DownloadQueueItem::FIELD_STATE_STR);
}

void DownloadQueueModel::onFileRemoved(QString collection_name, QString subject, quint64 total_size, quint64 total_size_remaining)
{
    DownloadQueueItem *collection_node;
    DownloadQueueItem *subject_node;

    collection_node = lookupCollectionItem(collection_name);
    subject_node = lookupSubjectItem(collection_node, subject);
    if (subject_node == NULL) {
        qWarning("onFileRemoved: Subject '%s' in collection '%s' could not be found", subject.toUtf8().data(), collection_name.toUtf8().data());
        return;
    }

    beginRemoveRows(createIndex(collection_node->row(), 0, collection_node), subject_node->row(), subject_node->row());

    collection_node->size = total_size;
    collection_node->sizeStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size);
    collection_node->sizeRemaining = total_size_remaining;
    collection_node->sizeRemainingStr = QNNTPGrabGlue::utilsCalculateFileSize(total_size_remaining);
    collection_node->progress = calculateProgress(total_size, total_size_remaining);

    collection_node->removeChild(subject_node);
    delete subject_node;

    endRemoveRows();

    QModelIndex idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, collection_node);
    emit dataChanged(idx, idx);

    idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_STR, collection_node);
    emit dataChanged(idx, idx);

    idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_FILE_SIZE_REMAINING_STR, collection_node);
    emit dataChanged(idx, idx);
}

void DownloadQueueModel::onFileMoved(QString orig_collection_name, QString subject, QString new_collection_name, int old_position, int new_position)
{
    Q_UNUSED(subject);
    Q_UNUSED(new_collection_name);

#if QT_VERSION < 0x040600
    static bool beenhere = false;
    if (!beenhere) {
        beenhere = true;
        qWarning("Moving files requires a more recent version of Qt. Please upgrade");
        return;
    }
#else
    DownloadQueueItem *collection_node = lookupCollectionItem(orig_collection_name);
    DownloadQueueItem *subject_node = lookupSubjectItem(collection_node, subject);

    qDebug("file moved: old_position: %i, new position: %i", old_position, new_position);

    if (collection_node == NULL) {
        qWarning("onFileMoved: Collection '%s' not found", orig_collection_name.toUtf8().data());
        return;
    }

    /* Do we need to move the item to the bottom */
    if (new_position == -1) {
        new_position = collection_node->childCount();
    }

    QModelIndex parent_idx = createIndex(subject_node->row(), 0, collection_node);
    int new_pos_in_model = qMin(collection_node->childCount(), ((old_position < new_position) ? new_position + 1 : new_position));
    if (!beginMoveRows(parent_idx, old_position, old_position, parent_idx, new_pos_in_model)) {
        qWarning("beginMoveRows failed\n");
        return;
    }

    collection_node->removeChild(subject_node);
    collection_node->insertChild(new_position, subject_node);

    endMoveRows();
#endif
}

void DownloadQueueModel::onCollectionAdded(QString collection_name, QString poster)
{
    beginInsertRows(QModelIndex(), rootItem->childCount(), rootItem->childCount());

    DownloadQueueItem *collection_node = new DownloadQueueItem(rootItem);
    collection_node->subject = collection_name;
    collection_node->poster = poster;

    rootItem->appendChild(collection_node);

    endInsertRows();

    QModelIndex idx_begin = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_STATE_STR, collection_node);
    QModelIndex idx_end = createIndex(collection_node->row(), (int) DownloadQueueItem::NUM_FIELDS, collection_node);
    emit dataChanged(idx_begin, idx_end);

    emit requestColumnResize(DownloadQueueItem::FIELD_SUBJECT);
}

void DownloadQueueModel::onCollectionRemoved(QString collection_name)
{
    DownloadQueueItem *collection_node = lookupCollectionItem(collection_name);

    if (collection_node == NULL) {
        qWarning("collectionmodified: Collection %s not found", collection_name.toUtf8().data());
        return;
    }

    /* Does the collection still contain files? */
    if (collection_node->childCount() != 0) {
        qWarning("Request received to remove collection, but collection is not empty");
        return;
    }

    beginRemoveRows(createIndex(rootItem->row(), 0, rootItem), collection_node->row(), collection_node->row());

    rootItem->removeChild(collection_node);
    //delete collection_node;       /* Disabled for now as this causes a segfault later on */

    endRemoveRows();
}

void DownloadQueueModel::onCollectionModified(QString collection_name, QString poster)
{
    DownloadQueueItem *collection_node = lookupCollectionItem(collection_name);

    if (collection_node == NULL) {
        qWarning("collectionmodified: Collection %s not found", collection_name.toUtf8().data());
        return;
    }

    collection_node->poster = poster;

    QModelIndex idx = createIndex(collection_node->row(), (int) DownloadQueueItem::FIELD_POSTER, collection_node);
    emit dataChanged(idx, idx);
}

void DownloadQueueModel::onCollectionMoved(QString collection_name, int old_position, int new_position)
{
#if QT_VERSION < 0x040600
    static bool beenhere = false;
    if (!beenhere) {
        beenhere = true;
        qWarning("Moving collections requires a more recent version of Qt. Please upgrade");
        return;
    }
#else
#if 0
    bool was_expanded;
#endif
    DownloadQueueItem *collection_node;

#if 0
    was_expanded = ui->treeWidget->topLevelItem(old_position)->isExpanded();
#endif
    collection_node = lookupCollectionItem(collection_name);

    /* Do we need to move the item to the bottom */
    if (new_position == -1) {
        new_position = rootItem->childCount();
    }

    QModelIndex parent_idx = createIndex(rootItem->row(), 0, rootItem);
    int new_pos_in_model = qMin(rootItem->childCount(), ((old_position < new_position) ? new_position + 1 : new_position));
    if (!beginMoveRows(QModelIndex(), old_position, old_position, QModelIndex(), new_pos_in_model)) {
        qWarning("beginMoveRows failed\n");
        return;
    }

    rootItem->removeChild(collection_node);
    rootItem->insertChild(new_position, collection_node);

    endMoveRows();

#if 0
    if (was_expanded) {
        collection_node->setExpanded(true);
    }
#endif
#endif
}

void DownloadQueueModel::onTrafficMonitorUpdate(int bytes_received1, int bytes_received2, int bytes_received3, int bytes_received4, int bytes_received5, int bytes_received6, int bytes_received7, int bytes_received8, int bytes_received9, int bytes_received10, QDateTime stamp, double average)
{
    Q_UNUSED(stamp);
    Q_UNUSED(average);

    static int times_been_here = 10;
    bool no_data_received = FALSE;
    quint64 total_estimated_time_remaining;

    if (bytes_received1 == 0 && bytes_received2 == 0 && bytes_received3 == 0 && bytes_received4 == 0 && bytes_received5 == 0 &&
        bytes_received6 == 0 && bytes_received7 == 0 && bytes_received8 == 0 && bytes_received9 == 0 && bytes_received10 == 0) {
            no_data_received = TRUE;
    }

    /* Only update on startup, on every 3 seconds, and when there has no data been received in the last 10 seconds */
    if (times_been_here < 3 && !no_data_received) {
        times_been_here += 1;
        return;
    }

    if (no_data_received) {
        times_been_here = 3;    /* update the time remaining the next time we get here */
    } else {
        times_been_here = 0;    /* perform the next update in 10 seconds */
    }

    total_estimated_time_remaining = 0;

    QModelIndex idx;
    int count = rootItem->childCount();
    for (int i = 0; i < count; i++) {
        DownloadQueueItem *collection_item;
        int estimated_time_remaining;
        quint64 file_size;

        collection_item = rootItem->child(i);
        file_size = collection_item->sizeRemaining;
        estimated_time_remaining = QNNTPGrabGlue::utilsCalculateEstimatedTimeRemaining(bytes_received1, bytes_received2, bytes_received3,
                                                                                       bytes_received4, bytes_received5, bytes_received6,
                                                                                       bytes_received7, bytes_received8, bytes_received9,
                                                                                       bytes_received10, file_size);

        if (estimated_time_remaining == -1) {
            /* Not enough data received to make an estimation */
            return;
        }

        if (estimated_time_remaining == 0) {
            /* Collection is already downloaded */
            if (!collection_item->estimatedTimeRemainingStr.isEmpty()) {
                collection_item->estimatedTimeRemainingStr = "";
                collection_item->estimatedTimeToFinishStr = "";

                idx = createIndex(collection_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_REMAINING_STR, collection_item);
                emit dataChanged(idx, idx);

                idx = createIndex(collection_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_TO_FINISH_STR, collection_item);
                emit dataChanged(idx, idx);
            }
        } else {
            collection_item->estimatedTimeRemainingStr = QNNTPGrabGlue::utilsGetReadableTimeRemaining(total_estimated_time_remaining + estimated_time_remaining);
            collection_item->estimatedTimeToFinishStr = QNNTPGrabGlue::utilsGetReadableFinishedTime(total_estimated_time_remaining + estimated_time_remaining);

            idx = createIndex(collection_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_REMAINING_STR, collection_item);
            emit dataChanged(idx, idx);

            idx = createIndex(collection_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_TO_FINISH_STR, collection_item);
            emit dataChanged(idx, idx);
        }

        total_estimated_time_remaining += estimated_time_remaining;

#if 0
        /* This loop is still too much for Qt... */

        /* Walk through all the children */
        int count2 = collection_item->childCount();
        for (int j = 0; j < count2; j++) {
            DownloadQueueItem *subject_item = collection_item->child(j);

            file_size = subject_item->sizeRemaining;
            if (file_size == 0) {
                /* File is already downloaded */
                if (!subject_item->estimatedTimeRemainingStr.isEmpty()) {
                    subject_item->estimatedTimeRemainingStr = "";
                    subject_item->estimatedTimeToFinishStr = "";

                    idx = createIndex(subject_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_REMAINING_STR, subject_item);
                    emit dataChanged(idx, idx);

                    idx = createIndex(subject_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_TO_FINISH_STR, subject_item);
                    emit dataChanged(idx, idx);
                }
            } else {
                estimated_time_remaining = QNNTPGrabGlue::utilsCalculateEstimatedTimeRemaining(bytes_received1, bytes_received2, bytes_received3,
                                                                                               bytes_received4, bytes_received5, bytes_received6,
                                                                                               bytes_received7, bytes_received8, bytes_received9,
                                                                                               bytes_received10, file_size);

                subject_item->estimatedTimeRemainingStr = QNNTPGrabGlue::utilsGetReadableTimeRemaining(total_estimated_time_remaining);
                subject_item->estimatedTimeToFinishStr = QNNTPGrabGlue::utilsGetReadableFinishedTime(total_estimated_time_remaining);

                idx = createIndex(subject_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_REMAINING_STR, subject_item);
                emit dataChanged(idx, idx);

                idx = createIndex(collection_item->row(), (int) DownloadQueueItem::FIELD_ESTIMATED_TIME_TO_FINISH_STR, subject_item);
                emit dataChanged(idx, idx);
            }
        }
#endif
    }
}
