#include <QFileInfo>
#include "WidgetRepairAndUnpack.h"
#include "ui_WidgetRepairAndUnpack.h"

enum {
    FIELD_FILENAME,
    FIELD_ACTION_STR,
    FIELD_PROGRESS_STR,
    FIELD_COLLECTION_NAME,
    FIELD_ACTION_INT,
    FIELD_PAR2_OR_UNPACK_FILENAME,
    FIELD_PROGRESS
};

enum {
    ACTION_PAR2,
    ACTION_UNPACK
};

WidgetRepairAndUnpack::WidgetRepairAndUnpack(QNNTPGrabGlue *glue, QWidget *parent) :
    QWidget(parent),
    ui(new Ui::WidgetRepairAndUnpack)
{
    ui->setupUi(this);

    this->glue = glue;
    this->opts = glue->configGetOpts();

    ui->treeWidget->header()->setResizeMode(QHeaderView::ResizeToContents);

    progressbarDelegate = new ProgressBarDelegatePAR2AndUnpack(this);
    ui->treeWidget->setItemDelegateForColumn(FIELD_PROGRESS_STR, progressbarDelegate);

    ui->treeWidget->setColumnHidden(FIELD_COLLECTION_NAME, true);
    ui->treeWidget->setColumnHidden(FIELD_ACTION_INT, true);
    ui->treeWidget->setColumnHidden(FIELD_PAR2_OR_UNPACK_FILENAME, true);
    ui->treeWidget->setColumnHidden(FIELD_PROGRESS, true);

    connect(glue, SIGNAL(configChanged()), SLOT(configChanged()));
    connect(glue, SIGNAL(pluginEvent(QString,QString,QList<QString>)), SLOT(pluginEvent(QString,QString,QList<QString>)));
}

WidgetRepairAndUnpack::~WidgetRepairAndUnpack()
{
    delete ui;
}

void WidgetRepairAndUnpack::changeEvent(QEvent *e)
{
    QWidget::changeEvent(e);
    switch (e->type()) {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

QTreeWidgetItem *WidgetRepairAndUnpack::lookupCollection(QString collection_name)
{
    for (int i = 0; i < ui->treeWidget->topLevelItemCount(); i++) {
        QTreeWidgetItem *item = ui->treeWidget->topLevelItem(i);

        if (item->text(FIELD_COLLECTION_NAME) == collection_name) {
            return item;
        }
    }

    return NULL;
}

QTreeWidgetItem *WidgetRepairAndUnpack::lookupFilename(QTreeWidgetItem *collection_item, QString filename)
{
    if (!collection_item) {
        return NULL;
    }

    int count = collection_item->childCount();
    for (int i = 0; i < count; i++) {
        QTreeWidgetItem *item = collection_item->child(i);

        if (item->text(FIELD_FILENAME) == filename) {
            return item;
        }
    }

    return NULL;
}

void WidgetRepairAndUnpack::lookupCollectionAndFilename(QString collection_name, QString filename, QTreeWidgetItem **collection_item, QTreeWidgetItem **file_item)
{
    QFileInfo fi(filename);
    QString basename = fi.fileName();
    QTreeWidgetItem *collection_item_tmp = lookupCollection(collection_name);

    if (!collection_item_tmp) {
        par2BeginVerify(collection_name, filename);
        collection_item_tmp = lookupCollection(collection_name);
        Q_ASSERT(collection_item_tmp != NULL);
    }

    QTreeWidgetItem *file_item_tmp = lookupFilename(collection_item_tmp, basename);
    if (!file_item_tmp) {
        par2BeginVerify(collection_name, filename);
        file_item_tmp = lookupFilename(collection_item_tmp, basename);
        Q_ASSERT(file_item_tmp != NULL);
    }

    *collection_item = collection_item_tmp;
    *file_item = file_item_tmp;

    Q_ASSERT(*collection_item != NULL);
    Q_ASSERT(*file_item != NULL);
}

void WidgetRepairAndUnpack::par2BeginVerify(QString collection_name, QString active_par2_filename)
{
    QFileInfo fi(active_par2_filename);
    QString filename = fi.fileName();

    QTreeWidgetItem *collection_item = lookupCollection(collection_name);
    if (!collection_item) {
        /* Collection not found yet, create a new entry */
        collection_item = new QTreeWidgetItem();
        collection_item->setText(FIELD_COLLECTION_NAME, collection_name);
        collection_item->setText(FIELD_FILENAME, collection_name);
        ui->treeWidget->addTopLevelItem(collection_item);
        ui->treeWidget->expandAll();
    }

    QTreeWidgetItem *file_item = lookupFilename(collection_item, filename);
    if (!file_item) {
        /* PAR2 set was already found, clear all the previous contents */
        delete file_item;
        file_item = NULL;
    }

    file_item = new QTreeWidgetItem(collection_item);
    file_item->setText(FIELD_FILENAME, filename);
    file_item->setText(FIELD_PAR2_OR_UNPACK_FILENAME, filename);
    if (fi.suffix().compare("par2", Qt::CaseInsensitive) == 0) {
        file_item->setText(FIELD_ACTION_INT, QString("%1").arg(ACTION_PAR2));
        file_item->setText(FIELD_ACTION_STR, tr("PAR2 repair"));
        file_item->setText(FIELD_PROGRESS_STR, tr("Now verifying files"));
        collection_item->setText(FIELD_PROGRESS_STR, tr("Now verifying files"));
    } else {
        file_item->setText(FIELD_ACTION_INT, QString("%1").arg(ACTION_UNPACK));
        file_item->setText(FIELD_ACTION_STR, tr("Unpack"));
        file_item->setText(FIELD_PROGRESS_STR, tr("Now unpacking files"));
        collection_item->setText(FIELD_PROGRESS_STR, tr("Now unpacking files"));
    }
}

void WidgetRepairAndUnpack::par2LoadProgressUpdate(QString collection_name, QString active_par2_filename, QString filename, double progress)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *par2_file_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, active_par2_filename, &collection_item, &par2_file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::par2LoadProgressUpdate: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!par2_file_item) {
        qWarning("WidgetRepairAndUnpack::par2LoadProgressUpdate: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!(file_item = lookupFilename(par2_file_item, filename))) {
        file_item = new QTreeWidgetItem(par2_file_item);
        file_item->setText(FIELD_FILENAME, filename);
        ui->treeWidget->expandAll();
    }
    file_item->setText(FIELD_PROGRESS, QString("%1").arg(progress));
    file_item->setText(FIELD_PROGRESS_STR, file_item->text(FIELD_PROGRESS_STR));    /* Force redraw */
}

void WidgetRepairAndUnpack::par2RecoveryFileLoaded(QString collection_name, QString active_par2_filename, QString filename, int num_new_packets, int num_blocks_found)
{
    Q_UNUSED(num_blocks_found);

    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *par2_file_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, active_par2_filename, &collection_item, &par2_file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::par2RecoveryFileLoaded: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!par2_file_item) {
        qWarning("WidgetRepairAndUnpack::par2RecoveryFileLoaded: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!(file_item = lookupFilename(par2_file_item, filename))) {
        file_item = new QTreeWidgetItem(par2_file_item);
        file_item->setText(FIELD_FILENAME, filename);
        ui->treeWidget->expandAll();
    }
    file_item->setText(FIELD_PROGRESS, "100");
    file_item->setText(FIELD_PROGRESS_STR, tr("%1 new PAR2 recovery blocks found").arg(num_new_packets));
    ui->treeWidget->scrollToBottom();
}

void WidgetRepairAndUnpack::par2FileLoaded(QString collection_name, QString active_par2_filename, QString filename, QString state, int num_blocks_found, int num_blocks_expected)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *par2_file_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, active_par2_filename, &collection_item, &par2_file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::par2FileLoaded: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!par2_file_item) {
        qWarning("WidgetRepairAndUnpack::par2FileLoaded: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    int progress = 0;
    QString msg;
    if (state == "MISSING") {
        progress = 0;
        msg = tr("File is missing");
    } else if (state == "NO_NEW_BLOCKS_FOUND") {
        progress = 100;
        msg = tr("No new blocks found");
    } else if (num_blocks_expected == num_blocks_found) {
        progress = 100;
        msg = tr("File is complete");
    } else {
        double progress_tmp = num_blocks_found;
        progress_tmp /= num_blocks_expected;
        progress_tmp *= 100.0;
        progress = (int) progress_tmp;
        msg = tr("%1 blocks found (expected: %2)").arg(num_blocks_found).arg(num_blocks_expected);
    }

    if (!(file_item = lookupFilename(par2_file_item, filename))) {
        file_item = new QTreeWidgetItem(par2_file_item);
        file_item->setText(FIELD_FILENAME, filename);
        ui->treeWidget->expandAll();
    }
    file_item->setText(FIELD_PROGRESS, QString("%1").arg(progress));
    file_item->setText(FIELD_PROGRESS_STR, msg);
    ui->treeWidget->scrollToBottom();
}

void WidgetRepairAndUnpack::par2RepairProgressUpdate(QString collection_name, QString active_par2_filename, double progress)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, active_par2_filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::par2RepairProgressUpdate: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::par2RepairProgressUpdate: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    collection_item->setText(FIELD_PROGRESS, "25");
    file_item->setText(FIELD_PROGRESS, QString("%1").arg((int) progress));
    file_item->setText(FIELD_PROGRESS_STR, tr("Repairing"));
}

void WidgetRepairAndUnpack::par2RepairFailure(QString collection_name, QString active_par2_filename, QString message, int num_blocks_more_required)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, active_par2_filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::par2RepairFailure: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::par2RepairFailure: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    QString msg;
    if (num_blocks_more_required == -1) {
        msg = message;
    } else {
        msg = tr("PAR2 repair failed: %1 more blocks required").arg(num_blocks_more_required);
    }

    collection_item->setText(FIELD_PROGRESS, "100");
    collection_item->setText(FIELD_PROGRESS_STR, msg);
    file_item->setText(FIELD_PROGRESS, "100");
    file_item->setText(FIELD_PROGRESS_STR, msg);
}

void WidgetRepairAndUnpack::par2RepairSuccess(QString collection_name, QString active_par2_filename)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, active_par2_filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::par2RepairSuccess: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::par2RepairSuccess: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (opts.enable_auto_unpack) {
        collection_item->setText(FIELD_PROGRESS, "50");
    } else {
        collection_item->setText(FIELD_PROGRESS, "100");
    }
    collection_item->setText(FIELD_PROGRESS_STR, tr("PAR2 repair succeeded"));
    file_item->setText(FIELD_PROGRESS, "100");
    file_item->setText(FIELD_PROGRESS_STR, tr("PAR2 repair succeeded"));
}

void WidgetRepairAndUnpack::par2NoRepairRequired(QString collection_name, QString active_par2_filename)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, active_par2_filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::par2NoRepairRequired: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::par2NoRepairRequired: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), active_par2_filename.toUtf8().data());
        return;
    }

    if (opts.enable_auto_unpack) {
        collection_item->setText(FIELD_PROGRESS, "50");
    } else {
        collection_item->setText(FIELD_PROGRESS, "100");
    }
    collection_item->setText(FIELD_PROGRESS_STR, tr("No repair needed"));
    file_item->setText(FIELD_PROGRESS, "100");
    file_item->setText(FIELD_PROGRESS_STR, tr("No repair needed"));
}

void WidgetRepairAndUnpack::unpackProgressUpdate(QString collection_name, QString filename, double progress)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    if (progress < 0) {
        /* Qt doesn't support pulse progress bars..Ignore for now */
        return;
    }

    lookupCollectionAndFilename(collection_name, filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::unpackProgressUpdate: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::unpackProgressUpdate: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    collection_item->setText(FIELD_PROGRESS, "75");
    file_item->setText(FIELD_PROGRESS, QString("%1").arg((int) (progress * 100)));
}

void WidgetRepairAndUnpack::unpackMessageReceived(QString collection_name, QString filename, QString message)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::unpackMessageReceived: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::unpackMessageReceived: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    file_item->setText(FIELD_PROGRESS_STR, message);
    ui->treeWidget->scrollToBottom();
}

void WidgetRepairAndUnpack::unpackWorkingArchiveChanged(QString collection_name, QString filename, QString working_archive)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::unpackWorkingArchiveChanged: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::unpackWorkingArchiveChanged: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    file_item->setText(FIELD_PROGRESS_STR, tr("Now unpacking archive '%1'").arg(working_archive));
    ui->treeWidget->scrollToBottom();
}

void WidgetRepairAndUnpack::unpackSuccess(QString collection_name, QString filename)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::unpackSuccess: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::unpackSuccess: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    collection_item->setText(FIELD_PROGRESS, "100");
    collection_item->setText(FIELD_PROGRESS_STR, tr("Unpack completed"));
    file_item->setText(FIELD_PROGRESS, "100");
    file_item->setText(FIELD_PROGRESS_STR, tr("Unpack completed"));
}

void WidgetRepairAndUnpack::unpackFailure(QString collection_name, QString filename, QString reason)
{
    QTreeWidgetItem *collection_item = NULL;
    QTreeWidgetItem *file_item = NULL;

    lookupCollectionAndFilename(collection_name, filename, &collection_item, &file_item);

    if (!collection_item) {
        qWarning("WidgetRepairAndUnpack::unpackFailure: collection_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    if (!file_item) {
        qWarning("WidgetRepairAndUnpack::unpackFailure: file_item is NULL for collection_name '%s' and filename '%s'", collection_name.toUtf8().data(), filename.toUtf8().data());
        return;
    }

    collection_item->setText(FIELD_PROGRESS, "100");
    collection_item->setText(FIELD_PROGRESS_STR, tr("Unpack failed due to an error"));
    file_item->setText(FIELD_PROGRESS, "100");
    file_item->setText(FIELD_PROGRESS_STR, tr("Unpack failed due to an error: ") + reason);
}

void WidgetRepairAndUnpack::configChanged()
{
    this->opts = glue->configGetOpts();
}

void WidgetRepairAndUnpack::pluginEvent(QString plugin_name, QString event_name, QList<QString> values)
{
    if (plugin_name == "PAR2") {
        if (event_name == "par2_begin_verify") {
            par2BeginVerify(values[0], values[1]);
        } else if (event_name == "par2_load_progress_update") {
            par2LoadProgressUpdate(values[0], values[1], values[2], values[3].toDouble());
        } else if (event_name == "par2_recovery_file_loaded") {
            par2RecoveryFileLoaded(values[0], values[1], values[2], values[3].toInt(), values[4].toInt());
        } else if (event_name == "par2_file_loaded") {
            par2FileLoaded(values[0], values[1], values[2], values[3], values[4].toInt(), values[5].toInt());
        } else if (event_name == "par2_repair_progress_update") {
            par2RepairProgressUpdate(values[0], values[1], values[2].toDouble());
        } else if (event_name == "par2_repair_failure") {
            par2RepairFailure(values[0], values[1], values[2], values[3].toInt());
        } else if (event_name == "par2_repair_success") {
            par2RepairSuccess(values[0], values[1]);
        } else if (event_name == "par2_no_repair_required") {
            par2NoRepairRequired(values[0], values[1]);
        } else {
            qWarning("%s", QString("Unknown plugin event received from plugin '%1': %2").arg(plugin_name, event_name).toUtf8().data());
        }
    } else if (plugin_name == "Unpack") {
        if (event_name == "unpack_progress_update") {
            unpackProgressUpdate(values[0], values[1], values[2].toDouble());
        } else if (event_name == "unpack_message_received") {
            unpackMessageReceived(values[0], values[1], values[2]);
        } else if (event_name == "unpack_working_archive_changed") {
            unpackWorkingArchiveChanged(values[0], values[1], values[2]);
        } else if (event_name == "unpack_success") {
            unpackSuccess(values[0], values[1]);
        } else if (event_name == "unpack_failure") {
            unpackFailure(values[0], values[1], values[2]);
        } else {
            qWarning("%s", QString("Unknown plugin event received from plugin '%1': %2").arg(plugin_name, event_name).toUtf8().data());
        }
    }
}
